// *************************************************************************************
// Test the speed of the H.26L integer DCT and the binDCT.
// Jie Liang, 03/06/2001
// *************************************************************************************

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <stdlib.h>

typedef unsigned char byte;

#define BLOCK_SIZE 4

struct img_par
{
  int number;                  /* current image number to be encoded*/
  int current_mb_nr;
  int current_slice_nr;
  int type;
  int multframe_no;
  int qp;                      /* quant for the current frame */
  int frame_cycle;
  int framerate;
  int width;                   /* Number of pels */
  int width_cr;                /* Number of pels chroma */
  int height;                  /* Number of lines */
  int height_cr;               /* Number of lines  chroma */
  int height_err;              /* For y vector limitation for error robustness*/
  int mb_y;                    /* current MB vertical   */
  int mb_x;                    /* current MB horizontal */
  int block_y;                 /* current block vertical   */
  int block_x;                 /* current block horizontal */
  int pix_y;                   /* current pixel vertical   */
  int pix_x;                   /* current pixel horizontal */
  int mb_y_upd;
  int mb_y_intra;              /* which GOB to intra code */
  int mb_mode;                 /* MB mode relevant for inter images, for intra images are all MB intra*/
  int imod;                    /* new/old intra modes + inter */
  int pix_c_y;                 /* current pixel chroma vertical   */
  int block_c_x;               /* current block chroma vertical   */
  int pix_c_x;                 /* current pixel chroma horizontal   */
  int blc_size_h;              /* block size for motion search, horizontal */
  int blc_size_v;              /* block size for motion search, vertical */
  int kac;                     /* any AC coeffs */
  int spiral_search[2][6561];  /* To obtain spiral search */
  int ipredmode[90][74];       /* prediction mode for inter frames */ /* fix from ver 4.1 */
  int mprr[6][4][4];           /* all 5 prediction modes */
  int mprr_2[5][16][16];       /* all 4 new intra prediction modes */
  int mv[4][4][5][9][2];       /* motion vectors for all block types and all reference frames */
  int mpr[16][16];             /* current best prediction mode */
  int m7[16][16];              /* the diff pixel values between orginal image and prediction */
  int cof[4][6][18][2][2];     /* coefficients */
  int cofu[5][2][2];           /* coefficients chroma */
  int quad[256];               /* Array containing square values,used for snr computation  */                                         /* Values are limited to 5000 for pixel differences over 70 (sqr(5000)). */
  int mv_bituse[512];
  int slice_numbers[396];

  byte li[8];                  /* 8 pix input for new filter routine loop() */
  byte lu[6];                  /* 4 pix output for new filter routine loop() */

  int lpfilter[6][602];        /* for use in loopfilter */

};

int testimg[4][4] = {
	{52, 55, 61, 66},
	{63, 59, 66, 90},
	{62, 59, 68, 113},
	{63, 58, 71, 122}
};

void intdct(struct img_par *img);
void bindct(struct img_par *img);
int main (int argc, char **argv);



/**********************************
 * H.26L Int DCT
 **********************************/
void intdct(struct img_par *img) {
	int i,j, i1, j1, m5[4];

	/*  Horizontal transform */

	for (j=0; j < BLOCK_SIZE; j++)
	{
		for (i=0; i < 2; i++)
		{
			i1=3-i;
			m5[i]=img->m7[i][j]+img->m7[i1][j];
			m5[i1]=img->m7[i][j]-img->m7[i1][j];
		}
		img->m7[0][j]=(m5[0]+m5[1])*13;
		img->m7[2][j]=(m5[0]-m5[1])*13;
		img->m7[1][j]=m5[3]*17+m5[2]*7;
		img->m7[3][j]=m5[3]*7-m5[2]*17;
	}

	/*  Vertival transform */

	for (i=0; i < BLOCK_SIZE; i++)
	{
		for (j=0; j < 2; j++)
		{
			j1=3-j;
			m5[j]=img->m7[i][j]+img->m7[i][j1];
			m5[j1]=img->m7[i][j]-img->m7[i][j1];
		}
		img->m7[i][0]=(m5[0]+m5[1])*13;
		img->m7[i][2]=(m5[0]-m5[1])*13;
		img->m7[i][1]=m5[3]*17+m5[2]*7;
		img->m7[i][3]=m5[3]*7-m5[2]*17;
	}
}


/**********************************
 * binDCT
 **********************************/
void bindct(struct img_par *img) {
	int i,j, m5[4];

	/*  Horizontal transform */

	for (i=0; i < BLOCK_SIZE; i++)
	{
		m5[0]=img->m7[i][0] + img->m7[i][3];
		m5[3]=img->m7[i][0] - img->m7[i][3];
		m5[1]=img->m7[i][1] + img->m7[i][2];
		m5[2]=img->m7[i][1] - img->m7[i][2];

		img->m7[i][0] = m5[0] + m5[1];
		img->m7[i][2] = (img->m7[i][0] >> 1) - m5[1];
		img->m7[i][3] = (((m5[3] << 3) - m5[3]) >> 4) - m5[2];
		img->m7[i][1] = m5[3] - (((img->m7[i][3] << 1) + img->m7[i][3]) >> 3);
	}

	/*  Vertical transform */

	for (j=0; j < BLOCK_SIZE; j++)
	{
		m5[0] = img->m7[0][j] + img->m7[3][j];
		m5[3] = img->m7[0][j] - img->m7[3][j];
		m5[1] = img->m7[1][j] + img->m7[2][j];
		m5[2] = img->m7[1][j] - img->m7[2][j];
		
		img->m7[0][j] = m5[0] + m5[1];
		img->m7[2][j] = (img->m7[0][j] >> 1) - m5[1];
	    img->m7[3][j] = (((m5[3] << 3) - m5[3]) >> 4) - m5[2];
		img->m7[1][j] = m5[3] - (((img->m7[3][j] << 1) + img->m7[3][j]) >> 3);
	}
}

/*******************************
 * main
 *******************************/

int main (int argc, char **argv)
{

	int k, i, j, cfg, n;
	struct img_par    *img;        

	if (argc != 3) {
		fprintf(stderr,"usage: ./h26l_time cfg loops\n"
			"where:\n"
			"  cfg: 0 int DCT, 1 binDCT.\n"
			"  loops: number of loops for forward DCT timing test.\n"
			"Profile analysis:\n"
			"  gprof ./cjpeg_time | more\n");

		exit(-1);
	}

	img =  (struct img_par *) calloc(1, sizeof(struct img_par));


	n = atoi(argv[2]);
	cfg = atoi(argv[1]);

	for (k = 0; k < n; k++) {

	  for (i=0; i<4; i++){
		for (j=0; j<4; j++){
		  img->m7[i][j] = testimg[i][j];
		}
	  }
		switch(cfg) {
			case 0:
				intdct(img);
				break;
			case 1:
				bindct(img);
				break;
		}
	}

  //Display: for test only.
	
	for (i=0; i<4; i++){
		for (j=0; j<4; j++){
			printf("%d  ", img->m7[i][j]);
		}
	}
	printf("\n");

	return(0);
}
